#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>

#include <iomanip>
#include <iostream>

#include "wrap_ida.h"

#include "defun.h"
#include "error.h"
#include "gripes.h"
#include "oct-obj.h"
#include "ov-fcn.h"
#include "ov-cell.h"
#include "pager.h"
#include "unwind-prot.h"
#include "utils.h"
#include "variables.h"


// Global pointer for user defined function required by IDA.
static octave_function *ida_fcn;


ColumnVector
ida_user_function (const ColumnVector& x, const ColumnVector& xdot,
                     double t, octave_idx_type& ires)
{

ColumnVector retval;

  assert (x.capacity () == xdot.capacity ());

  octave_value_list args;

  args(2) = t;
  args(1) = xdot;
  args(0) = x;

  if (ida_fcn)
    {
      octave_value_list tmp = ida_fcn->do_multi_index_op (1, args);

      if (error_state)
        {
          gripe_user_supplied_eval ("ida");
          return retval;
        }

      int tlen = tmp.length ();
      if (tlen > 0 && tmp(0).is_defined ())
        {
          if (! warned_fcn_imaginary && tmp(0).is_complex_type ())
            {
              warning ("ida: ignoring imaginary part returned from user-supplied function");
              warned_fcn_imaginary = true;
            }

          retval = ColumnVector (tmp(0).vector_value ());

          if (tlen > 1)
            ires = tmp(1).int_value ();

          if (error_state || retval.length () == 0)
            gripe_user_supplied_eval ("ida");
        }
      else
        gripe_user_supplied_eval ("ida");
    }

  return retval;


}


DEFUN (ida, args, nargout, "INSERT DOCUMENTATION")
{
  octave_value_list retval;

  int nargin = args.length ();

  if (nargin > 3 && nargin < 6 && nargout < 5)
    {
      std::string fcn_name, fname, jac_name, jname;
      ida_fcn = 0;
      ida_jac = 0;

      octave_value f_arg = args(0);

      if (!ida_fcn && ! f_arg.is_cell ())
        {
          if (f_arg.is_function_handle () || f_arg.is_inline_function ())
            ida_fcn = f_arg.function_value ();
          else
	     //need error catch
	     return retval;
        }
      else
        //need error catch
	return retval;

      ColumnVector state = ColumnVector (args(1).vector_value ());
       // need error catching 

      ColumnVector deriv (args(2).vector_value ());
       // need error catching 

      ColumnVector out_times (args(3).vector_value ());
       // need error catching

      double rel_tol = 1e-4;
      double abs_tol = 1e-8;
      double tzero = out_times (0);

      DAEFunc func (ida_user_function);

      IDA dae (state, deriv, tzero, func);

      dae.set_options (rel_tol, abs_tol);

      Matrix output;
      Matrix deriv_output;

      output = dae.integrate (out_times, deriv_output);

      if (fcn_name.length ())
        clear_function (fcn_name);

      if (! error_state)
        {
          std::string msg = dae.error_message ();

          retval(3) = msg;
          retval(2) = static_cast<double> (dae.integration_state ());

          if (dae.integration_ok ())
            {
              retval(1) = deriv_output;
              retval(0) = output;
            }
          else
            {
              retval(1) = Matrix ();
              retval(0) = Matrix ();

              if (nargout < 3)
                error ("ida: %s", msg.c_str ());
            }
        }
    }
  else
    print_usage ();

  return retval;
}

